use std::sync::Arc;

use crate::{
    config::Config,
    error::{AppError, HandlerError, HandlerResult},
    host,
};

use socks5_server::{
    connection::state::NeedAuthenticate,
    proto::{Address, Reply},
    Command, IncomingConnection,
};
use tokio::{io::AsyncWriteExt, net::TcpStream};

pub async fn handler(
    conn: IncomingConnection<(), NeedAuthenticate>,
    config: Arc<Config>,
) -> HandlerResult<()> {
    let conn = conn.authenticate().await?.0;

    match conn.wait().await? {
        Command::Associate(cmd, _addr) => {
            let _ = cmd
                .reply(Reply::CommandNotSupported, Address::unspecified())
                .await?
                .close()
                .await;
        }

        Command::Bind(cmd, _addr) => {
            let _ = cmd
                .reply(Reply::CommandNotSupported, Address::unspecified())
                .await?
                .close()
                .await;
        }

        Command::Connect(cmd, addr) => {
            let target = match addr {
                Address::DomainAddress(host, port) => {
                    let mut addrs = match host::resolve_domain(host, port, config.clone()) {
                        Ok(iter) => iter,
                        Err(e) => {
                            let conn = cmd
                                .reply(Reply::GeneralFailure, Address::unspecified())
                                .await?
                                .into_inner();
                            return Err((e, conn).into());
                        }
                    };

                    TcpStream::connect(addrs.next().unwrap()).await // TODO
                }
                Address::SocketAddress(addr) => {
                    let addr = match host::resolve_ip(addr, config.clone()) {
                        Ok(addr) => addr,
                        Err(e) => {
                            let conn = cmd
                                .reply(Reply::GeneralFailure, Address::unspecified())
                                .await?
                                .into_inner();
                            return Err((e, conn).into());
                        }
                    };

                    TcpStream::connect(addr).await
                }
            };

            let mut target = match target {
                Ok(strm) => strm,
                Err(e) => {
                    let conn = cmd
                        .reply(Reply::HostUnreachable, Address::unspecified())
                        .await?
                        .into_inner();
                    return Err((e, conn).into());
                }
            };

            let mut conn = cmd.reply(Reply::Succeeded, Address::unspecified()).await?;

            let res = tokio::io::copy_bidirectional(&mut target, &mut conn).await;
            let _ = conn.shutdown().await;
            let _ = target.shutdown().await;

            // println!("2nd sd: {:?}", conn.shutdown().await);

            res.map_err(|e| (e, conn.into_inner()))?;
        }
    }

    Ok(())
}
