use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, ToSocketAddrs};

use serde::{de::Visitor, Deserialize};

pub const LOCALHOST_V4: IpAddr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
pub const DEFAULT_PORT: u16 = 5232;

#[derive(Debug, Clone, Eq, PartialEq, Hash)]
pub enum TargetAddr {
    DomainName(String, u16),
    IpAddress(SocketAddr),
}

impl<'de> Deserialize<'de> for TargetAddr {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        deserializer.deserialize_str(AddrVisitor)
    }
}

impl ToSocketAddrs for TargetAddr {
    type Iter = AddrIter;

    fn to_socket_addrs(&self) -> std::io::Result<Self::Iter> {
        Ok(match self {
            Self::DomainName(host, port) => (host.as_str(), *port).to_socket_addrs()?.into(),
            Self::IpAddress(sockaddr) => sockaddr.clone().into(),
        })
    }
}

impl From<(IpAddr, u16)> for TargetAddr {
    #[inline]
    fn from((addr, port): (IpAddr, u16)) -> Self {
        Self::IpAddress(SocketAddr::new(addr, port))
    }
}

pub struct AddrVisitor;

impl<'de> Visitor<'de> for AddrVisitor {
    type Value = TargetAddr;

    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        let v = v.trim();

        let ipv6 = v.starts_with('[');
        let mut port_idx = v.len();

        let port = if ipv6 {
            v.rfind("]:").and_then(|pos| {
                port_idx = pos;
                v.get(pos + 2..)
            })
        } else {
            v.rfind(':').and_then(|pos| {
                port_idx = pos;
                v.get(pos + 1..)
            })
        }
        .and_then(|port| {
            port.parse::<u16>()
                .map_err(|_| {
                    port_idx = v.len();
                })
                .ok()
        })
        .unwrap_or(DEFAULT_PORT);

        let host = if ipv6 {
            v.get(1..port_idx)
        } else {
            v.get(..port_idx)
        }
        .ok_or(serde::de::Error::missing_field("host"))?;

        if ipv6 {
            let addr = IpAddr::V6(host.parse().map_err(serde::de::Error::custom)?);
            Ok((addr, port).into())
        } else {
            if let Ok(addr) = host.parse::<Ipv4Addr>() {
                Ok((IpAddr::V4(addr), port).into())
            } else if let Ok(addr) = host.parse::<Ipv6Addr>() {
                Ok((IpAddr::V6(addr), port).into())
            } else {
                Ok(TargetAddr::DomainName(host.to_owned(), port))
            }
        }
    }

    fn expecting(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            concat!(
                "host:port ",
                "where host is a domain or an ipv4/ipv6 addr, ",
                "and optional port is u16",
            ),
        )
    }
}

pub enum AddrIter {
    HostResolved(std::vec::IntoIter<SocketAddr>),
    OneAddress(std::iter::Once<SocketAddr>),
}

impl Iterator for AddrIter {
    type Item = SocketAddr;

    #[inline]
    fn next(&mut self) -> Option<Self::Item> {
        match self {
            Self::HostResolved(iter) => iter.next(),
            Self::OneAddress(iter) => iter.next(),
        }
    }
}

impl From<std::vec::IntoIter<SocketAddr>> for AddrIter {
    #[inline]
    fn from(value: std::vec::IntoIter<SocketAddr>) -> Self {
        Self::HostResolved(value)
    }
}

impl From<SocketAddr> for AddrIter {
    #[inline]
    fn from(value: SocketAddr) -> Self {
        Self::OneAddress(std::iter::once(value))
    }
}

#[cfg(test)]
mod tests {
    use std::net::{IpAddr, Ipv6Addr};

    use serde_test::{assert_de_tokens, Token};

    use super::{TargetAddr, DEFAULT_PORT, LOCALHOST_V4};

    const TEST_V6: IpAddr = IpAddr::V6(Ipv6Addr::new(
        0xfe80, 0x0, 0x0, 0x0, 0x721e, 0xd21f, 0x29a3, 0xf396,
    ));

    #[test]
    pub fn ipv4_without_port() {
        let exp: TargetAddr = (LOCALHOST_V4, DEFAULT_PORT).into();
        assert_de_tokens(&exp, &[Token::Str("127.0.0.1")]);
    }

    #[test]
    pub fn ipv4_with_port() {
        let exp: TargetAddr = (LOCALHOST_V4, 443).into();
        assert_de_tokens(&exp, &[Token::Str("127.0.0.1:443")]);
    }

    #[test]
    pub fn ipv6_without_port() {
        let exp: TargetAddr = (TEST_V6, DEFAULT_PORT).into();
        assert_de_tokens(&exp, &[Token::Str("fe80::721e:d21f:29a3:f396")]);
    }

    #[test]
    pub fn ipv6_with_port() {
        let exp: TargetAddr = (TEST_V6, 443).into();
        assert_de_tokens(&exp, &[Token::Str("[fe80::721e:d21f:29a3:f396]:443")]);
    }

    #[test]
    pub fn domain_without_port() {
        let exp = TargetAddr::DomainName("smp.dc09.ru".to_owned(), DEFAULT_PORT);
        assert_de_tokens(&exp, &[Token::Str("smp.dc09.ru")]);
    }

    #[test]
    pub fn domain_with_port() {
        let exp = TargetAddr::DomainName("xftp.dc09.ru".to_owned(), 443);
        assert_de_tokens(&exp, &[Token::Str("xftp.dc09.ru:443")]);
    }
}
